from __future__ import with_statement
from __future__ import division

import traceback, itertools, sys, json, math, os

import mercantile
import rasterio
import numpy as np
import sqlite3
from multiprocessing import Pool
from rasterio._io import virtual_file_to_buffer
from riomucho.single_process_pool import MockTub

from io import BytesIO
from PIL import Image

from rasterio import transform
from rasterio.warp import reproject, transform_bounds
try:
    from rasterio.warp import RESAMPLING
except ImportError:
    from rasterio.enums import Resampling as RESAMPLING

from rio_rgbify.encoders import data_to_rgb

buffer = bytes if sys.version_info > (3,) else buffer

work_func = None
global_args = None
src = None


def _main_worker(inpath, g_work_func, g_args):
    """
    Util for setting global vars w/ a Pool
    """
    global work_func
    global global_args
    global src
    work_func = g_work_func
    global_args = g_args

    src = rasterio.open(inpath)


def _encode_as_webp(data, profile=None, affine=None):
    """
    Uses BytesIO + PIL to encode a (3, 512, 512)
    array into a webp bytearray.

    Parameters
    -----------
    data: ndarray
        (3 x 512 x 512) uint8 RGB array
    profile: None
        ignored
    affine: None
        ignored

    Returns
    --------
    contents: bytearray
        webp-encoded bytearray of the provided input data
    """
    with BytesIO() as f:
        im = Image.fromarray(np.rollaxis(data, 0, 3))
        im.save(f, format='webp', lossless=True)

        return f.getvalue()


def _encode_as_png(data, profile, dst_transform):
    """
    Uses rasterio's virtual file system to encode a (3, 512, 512)
    array as a png-encoded bytearray.

    Parameters
    -----------
    data: ndarray
        (3 x 512 x 512) uint8 RGB array
    profile: dictionary
        dictionary of kwargs for png writing
    affine: Affine
        affine transform for output tile

    Returns
    --------
    contents: bytearray
        png-encoded bytearray of the provided input data
    """
    profile['affine'] = dst_transform

    with rasterio.open('/vsimem/tileimg', 'w', **profile) as dst:
        dst.write(data)

    contents = bytearray(virtual_file_to_buffer('/vsimem/tileimg'))

    return contents


def _tile_worker(tile):
    """
    For each tile, and given an open rasterio src, plus a`global_args` dictionary
    with attributes of `base_val`, `interval`, and a `writer_func`,
    warp a continous single band raster to a 512 x 512 mercator tile,
    then encode this tile into RGB.

    Parameters
    -----------
    tile: list
        [x, y, z] indices of tile

    Returns
    --------
    tile, buffer
        tuple with the input tile, and a bytearray with the data encoded into
        the format created in the `writer_func`

    """
    x, y, z = tile

    bounds = [c for i in (mercantile.xy(*mercantile.ul(x, y + 1, z)),
            mercantile.xy(*mercantile.ul(x + 1, y, z))) for c in i]

    toaffine = transform.from_bounds(*bounds + [512, 512])

    out = np.empty((512, 512), dtype=src.meta['dtype'])

    reproject(
        rasterio.band(src, 1), out,
        dst_transform=toaffine,
        dst_crs="init='epsg:3857'",
        resampling=RESAMPLING.bilinear)

    out = data_to_rgb(out, global_args['base_val'], global_args['interval'])

    return tile, global_args['writer_func'](out, global_args['kwargs'].copy(), toaffine)


def _tile_range(min_tile, max_tile):
    """
    Given a min and max tile, return an iterator of
    all combinations of this tile range

    Parameters
    -----------
    min_tile: list
        [x, y, z] of minimun tile
    max_tile:
        [x, y, z] of minimun tile

    Returns
    --------
    tiles: iterator
        iterator of [x, y, z] tiles
    """
    min_x, min_y, _ = min_tile
    max_x, max_y, _ = max_tile

    return itertools.product(range(min_x, max_x + 1), range(min_y, max_y + 1))


def _make_tiles(bbox, src_crs, minz, maxz):
    '''
    Given a bounding box, zoom range, and source crs,
    find all tiles that would intersect

    Parameters
    -----------
    bbox: list
        [w, s, e, n] bounds
    src_crs: str
        the source crs of the input bbox
    minz: int
        minumum zoom to find tiles for
    maxz: int
        maximum zoom to find tiles for

    Returns
    --------
    tiles: generator
        generator of [x, y, z] tiles that intersect
        the provided bounding box
    '''
    w, s, e, n = transform_bounds(*[src_crs, 'epsg:4326'] + bbox, densify_pts=0)

    EPSILON = 1.0e-10

    w += EPSILON
    s += EPSILON
    e -= EPSILON
    n -= EPSILON

    for z in range(minz, maxz + 1):
        for x, y in _tile_range(
            mercantile.tile(w, n, z),
            mercantile.tile(e, s, z)):

            yield [x, y, z]


class RGBTiler:
    '''
    Takes continous source data of an arbitrary bit depth and encodes it
    in parallel into RGB tiles in an MBTiles file. Provided with a context manager:
    ```
    with RGBTiler(inpath, outpath, min_z, max_x, **kwargs) as tiler:
        tiler.run(processes)
    ```

    Parameters
    -----------
    inpath: string
        filepath of the source file to read and encode
    outpath: string
        filepath of the output `mbtiles`
    min_z: int
        minimum zoom level to tile
    max_z: int
        maximum zoom level to tile

    Keyword Arguments
    ------------------
    baseval: float
        the base value of the RGB numbering system.
        (will be treated as zero for this encoding)
        Default=0
    interval: float
        the interval at which to encode
        Default=1
    format: str
        output tile image format (png or webp)
        Default=png
    bounding_tile: list
        [x, y, z] of bounding tile; limits tiled output to this extent

    Returns
    --------
    None

    '''
    def __init__(self, inpath, outpath, min_z, max_z, interval=1, base_val=0, bounding_tile=None, **kwargs):
        self.run_function = _tile_worker
        self.inpath = inpath
        self.outpath = outpath
        self.min_z = min_z
        self.max_z = max_z
        self.bounding_tile = bounding_tile

        if not 'format' in kwargs:
            writer_func = _encode_as_png
            self.image_format = 'png'
        elif kwargs['format'].lower() == 'png':
            writer_func = _encode_as_png
            self.image_format = 'png'
        elif kwargs['format'].lower() == 'webp':
            writer_func = _encode_as_webp
            self.image_format = 'webp'
        else:
            raise ValueError('{0} is not a supported filetype!'.format(kwargs['format']))

        # global kwargs not used if output  is webp
        self.global_args = {
            'kwargs':  {
                'driver': 'PNG',
                'dtype': 'uint8',
                'height': 512,
                'width': 512,
                'count': 3,
                'crs': 'EPSG:3857'
            },
            'base_val': base_val,
            'interval': interval,
            'writer_func': writer_func
        }

    def __enter__(self):
        return self
    def __exit__(self, ext_t, ext_v, trace):
        if ext_t:
            traceback.print_exc()


    def run(self, processes=4):
        '''
        Warp, encode, and tile
        '''

        # get the bounding box + crs of the file to tile
        with rasterio.open(self.inpath) as src:
            bbox = list(src.bounds)
            src_crs = src.crs


        # remove the output filepath if it exists
        if os.path.exists(self.outpath):
            os.unlink(self.outpath)

        # create a connection to the mbtiles file
        conn = sqlite3.connect(self.outpath)
        cur = conn.cursor()

        # create the tiles table
        cur.execute(
            "CREATE TABLE tiles "
            "(zoom_level integer, tile_column integer, "
            "tile_row integer, tile_data blob);")
        # create empty metadata
        cur.execute(
            "CREATE TABLE metadata (name text, value text);")

        conn.commit()

        # populate metadata with required fields
        cur.execute(
            "INSERT INTO metadata "
            "(name, value) "
            "VALUES ('format', ?);",
            (self.image_format, ))

        cur.execute(
            "INSERT INTO metadata "
            "(name, value) "
            "VALUES ('name', '');")
        cur.execute(
            "INSERT INTO metadata "
            "(name, value) "
            "VALUES ('description', '');")
        cur.execute(
            "INSERT INTO metadata "
            "(name, value) "
            "VALUES ('version', '1');")
        cur.execute(
            "INSERT INTO metadata "
            "(name, value) "
            "VALUES ('type', 'baselayer');")

        conn.commit()

        if processes == 1:
            # use mock pool for profiling / debugging
            self.pool = MockTub(_main_worker, (self.inpath, self.run_function, self.global_args))
        else:
            self.pool = Pool(processes, _main_worker, (self.inpath, self.run_function, self.global_args))

        # generator of tiles to make
        if self.bounding_tile is None:
            tiles = _make_tiles(bbox, src_crs, self.min_z, self.max_z)
        else:
            constrained_bbox = list(mercantile.bounds(self.bounding_tile))
            tiles = _make_tiles(constrained_bbox, 'epsg:4326', self.min_z, self.max_z)


        for tile, contents in self.pool.imap_unordered(self.run_function,
                                                       tiles):
            x, y, z = tile

            # mbtiles use inverse y indexing
            tiley = int(math.pow(2, z)) - y - 1

            # insert tile object
            cur.execute(
                "INSERT INTO tiles "
                "(zoom_level, tile_column, tile_row, tile_data) "
                "VALUES (?, ?, ?, ?);",
                (z, x, tiley, buffer(contents)))

            conn.commit()

        conn.close()

        self.pool.close()
        self.pool.join()

        return None



import mercantile
import types

from hypothesis import given
import hypothesis.strategies as st
import pytest

import numpy as np
from rasterio import Affine
from rio_rgbify.mbtiler import (_encode_as_webp, _encode_as_png, _make_tiles, _tile_range, RGBTiler)


@given(
    st.integers(
        min_value=0, max_value=(2 ** 10 - 1)
        ),
    st.integers(
        min_value=0, max_value=(2 ** 10 - 1)))
def test_make_tiles_tile_bounds(x, y):
    '''
    Test if children tiles from z10 are created correctly
    '''
    test_bounds = mercantile.bounds(x, y, 10)

    test_bbox = list(mercantile.xy(test_bounds.west, test_bounds.south)) + list(mercantile.xy(test_bounds.east, test_bounds.north))

    test_crs = 'epsg:3857'
    test_minz = 10
    test_maxz = 13

    created_tiles_gen = _make_tiles(test_bbox, test_crs, test_minz, test_maxz)

    assert isinstance(created_tiles_gen, types.GeneratorType)

    created_tiles = list(created_tiles_gen)

    assert len(created_tiles) == 85


@given(
    st.lists(
        elements=st.integers(min_value=0, max_value=99),
        max_size=3, min_size=3
        ),
    st.lists(
        elements=st.integers(
        min_value=100, max_value=200
        ), max_size=3, min_size=3
        ))
def test_tile_range(mintile, maxtile):
    minx, miny, _ = mintile
    maxx, maxy, _ = maxtile

    expected_length = (maxx - minx + 1) * (maxy - miny + 1)
    assert expected_length == len(list(_tile_range(mintile, maxtile)))


def test_webp_writer():
    test_data = np.zeros((3, 256, 256), dtype=np.uint8)

    test_bytearray = _encode_as_webp(test_data)

    assert len(test_bytearray) == 42

    test_complex_data = test_data.copy()

    test_complex_data[0] += (np.random.rand(256, 256) * 255).astype(np.uint8)
    test_complex_data[1] += 10

    test_bytearray_complex = _encode_as_webp(test_complex_data)

    assert len(test_bytearray) < len(test_bytearray_complex)


def test_file_writer():
    test_data = np.zeros((3, 256, 256), dtype=np.uint8)

    test_opts = {
        'driver': 'PNG',
        'dtype': 'uint8',
        'height': 512,
        'width': 512,
        'count': 3,
        'crs': 'EPSG:3857'
    }

    test_affine = Affine(1, 0, 0, 0, -1, 0)

    test_bytearray = _encode_as_png(test_data, test_opts, test_affine)

    assert len(test_bytearray) == 842

    test_complex_data = test_data.copy()

    test_complex_data[0] += (np.random.rand(256, 256) * 255).astype(np.uint8)
    test_complex_data[1] += 10

    test_bytearray_complex = _encode_as_png(test_complex_data, test_opts, test_affine)

    assert len(test_bytearray) < len(test_bytearray_complex)


def test_webp_writer_fails_dtype():
    test_data = np.zeros((3, 256, 256), dtype=np.float64)

    with pytest.raises(TypeError):
        _encode_as_webp(test_data)


def test_png_writer_fails_dtype():
    test_data = np.zeros((3, 256, 256), dtype=np.float64)

    with pytest.raises(TypeError):
        _encode_as_png(test_data)


def test_RGBtiler_format_fails():
    test_in = 'i/do/not/exist.tif'
    test_out = 'nor/do/i.tif'
    test_minz = 0
    test_maxz = 1

    with pytest.raises(ValueError):
        with RGBTiler(test_in, test_out, test_minz, test_maxz,
            format='poo') as rtiler:
            pass



from click.testing import CliRunner
import rasterio as rio
import numpy as np
from rio_rgbify.scripts.cli import rgbify
import click

from tempfile import mkdtemp
from shutil import rmtree
import os

from raster_tester.compare import affaux, upsample_array

class TestingDir:
    def __init__(self):
        self.tmpdir = mkdtemp()
    def __enter__(self):
        return self
    def __exit__(self, a, b, c):
        rmtree(self.tmpdir)
    def mkpath(self, filename):
        return os.path.join(self.tmpdir, filename)

def flex_compare(r1, r2, thresh=10):
    upsample = 4
    r1 = r1[::upsample]
    r2 = r2[::upsample]
    toAff, frAff = affaux(upsample)
    r1 = upsample_array(r1, upsample, frAff, toAff)
    r2 = upsample_array(r2, upsample, frAff, toAff)
    tdiff = np.abs(r1.astype(np.float64) - r2.astype(np.float64))

    click.echo('{0} values exceed the threshold difference with a max variance of {1}'.format(
        np.sum(tdiff > thresh), tdiff.max()), err=True)

    return not np.any(tdiff > thresh)


def test_cli_good_elev():
    in_elev_src = 'test/fixtures/elev.tif'
    expected_src = 'test/expected/elev-rgb.tif'
    with TestingDir() as tmpdir:
        out_rgb_src = tmpdir.mkpath('rgb.tif')

        runner = CliRunner()
        result = runner.invoke(rgbify, [in_elev_src, out_rgb_src, '--interval', 0.001, '--base-val', -100, '-j', 1])

        assert result.exit_code == 0

        with rio.open(out_rgb_src) as created:
            with rio.open(expected_src) as expected:
                carr = created.read()
                earr = expected.read()
                for a, b in zip(carr, earr):
                    assert flex_compare(a, b)


def test_cli_fail_elev():
    in_elev_src = 'test/fixtures/elev.tif'
    expected_src = 'test/expected/elev-rgb.tif'
    with TestingDir() as tmpdir:
        out_rgb_src = tmpdir.mkpath('rgb.tif')

        runner = CliRunner()
        result = runner.invoke(rgbify, [in_elev_src, out_rgb_src, '--interval', 0.00000001, '--base-val', -100, '-j', 1])

        assert result.exit_code == -1


def test_mbtiler_webp():
    in_elev_src = 'test/fixtures/elev.tif'

    with TestingDir() as tmpdir:
        out_mbtiles_finer = tmpdir.mkpath('output-0-dot-1.mbtiles')
        runner = CliRunner()

        result_finer = runner.invoke(rgbify, [in_elev_src, out_mbtiles_finer, '--interval', 0.1, '--min-z', 10, '--max-z', 11, '--format', 'webp', '-j', 1])

        assert result_finer.exit_code == 0

        out_mbtiles_coarser  = tmpdir.mkpath('output-1.mbtiles')

        result_coarser = runner.invoke(rgbify, [in_elev_src, out_mbtiles_coarser, '--min-z', 10, '--max-z', 11, '--format', 'webp', '-j', 1])

        assert result_coarser.exit_code == 0

        assert os.path.getsize(out_mbtiles_finer) > os.path.getsize(out_mbtiles_coarser)


def test_mbtiler_png():
    in_elev_src = 'test/fixtures/elev.tif'

    with TestingDir() as tmpdir:
        out_mbtiles_finer = tmpdir.mkpath('output-0-dot-1.mbtiles')
        runner = CliRunner()

        result_finer = runner.invoke(rgbify, [in_elev_src, out_mbtiles_finer, '--interval', 0.1, '--min-z', 10, '--max-z', 11, '--format', 'png'])

        assert result_finer.exit_code == 0

        out_mbtiles_coarser  = tmpdir.mkpath('output-1.mbtiles')

        result_coarser = runner.invoke(rgbify, [in_elev_src, out_mbtiles_coarser, '--min-z', 10, '--max-z', 11, '--format', 'png', '-j', 1])

        assert result_coarser.exit_code == 0

        assert os.path.getsize(out_mbtiles_finer) > os.path.getsize(out_mbtiles_coarser)

def test_mbtiler_png_bounding_tile():
    in_elev_src = 'test/fixtures/elev.tif'

    with TestingDir() as tmpdir:
        out_mbtiles_not_limited = tmpdir.mkpath('output-not-limited.mbtiles')
        runner = CliRunner()

        result_not_limited = runner.invoke(rgbify, [in_elev_src, out_mbtiles_not_limited, '--min-z', 12, '--max-z', 12, '--format', 'png'])

        assert result_not_limited.exit_code == 0

        out_mbtiles_limited  = tmpdir.mkpath('output-limited.mbtiles')

        result_limited = runner.invoke(rgbify, [in_elev_src, out_mbtiles_limited, '--min-z', 12, '--max-z', 12, '--format', 'png', '--bounding-tile', '[654, 1582, 12]'])

        assert result_limited.exit_code == 0

        assert os.path.getsize(out_mbtiles_not_limited) > os.path.getsize(out_mbtiles_limited)


def test_mbtiler_webp_badzoom():
    in_elev_src = 'test/fixtures/elev.tif'

    with TestingDir() as tmpdir:
        out_mbtiles = tmpdir.mkpath('output.mbtiles')
        runner = CliRunner()

        result = runner.invoke(rgbify, [in_elev_src, out_mbtiles, '--min-z', 10, '--max-z', 9, '--format', 'webp', '-j', 1])

        assert result.exit_code == -1

def test_mbtiler_webp_badboundingtile():
    in_elev_src = 'test/fixtures/elev.tif'

    with TestingDir() as tmpdir:
        out_mbtiles = tmpdir.mkpath('output.mbtiles')
        runner = CliRunner()

        result = runner.invoke(rgbify, [in_elev_src, out_mbtiles, '--min-z', 10, '--max-z', 9, '--format', 'webp', '--bounding-tile', '654, 1582, 12'])

        assert result.exit_code == -1

def test_mbtiler_webp_badboundingtile_values():
    in_elev_src = 'test/fixtures/elev.tif'

    with TestingDir() as tmpdir:
        out_mbtiles = tmpdir.mkpath('output.mbtiles')
        runner = CliRunner()

        result = runner.invoke(rgbify, [in_elev_src, out_mbtiles, '--min-z', 10, '--max-z', 9, '--format', 'webp', '--bounding-tile', '[654, 1582]'])

        assert result.exit_code == -1


def test_bad_input_format():
    in_elev_src = 'test/fixtures/elev.tif'

    with TestingDir() as tmpdir:
        out_mbtiles = tmpdir.mkpath('output.lol')
        runner = CliRunner()

        result = runner.invoke(rgbify, [in_elev_src, out_mbtiles, '--min-z', 10, '--max-z', 9, '--format', 'webp', '-j', 1])

        assert result.exit_code == -1



from __future__ import division
import numpy as np

def data_to_rgb(data, baseval, interval):
    '''
    Given an arbitrary (rows x cols) ndarray,
    encode the data into uint8 RGB from an arbitrary
    base and interval

    Parameters
    -----------
    data: ndarray
        (rows x cols) ndarray of data to encode
    baseval: float
        the base value of the RGB numbering system.
        will be treated as zero for this encoding
    interval: float
        the interval at which to encode

    Returns
    --------
    ndarray: rgb data
        a uint8 (3 x rows x cols) ndarray with the
        data encoded
    '''
    data = data.astype(np.float64)
    data -= baseval
    data /= interval

    rows, cols = data.shape

    datarange = data.max() - data.min()

    if _range_check(datarange):
        raise ValueError('Data of {} larger than 256 ** 3'.format(datarange))

    rgb = np.zeros((3, rows, cols), dtype=np.uint8)

    rgb[2] = (((data / 256) - (data // 256)) * 256)
    rgb[1] = ((((data // 256) / 256) - ((data // 256) // 256)) * 256)
    rgb[0] = (((((data // 256) // 256) / 256) - (((data // 256) // 256) // 256)) * 256)

    return rgb


def _decode(data, base, interval):
    '''
    Utility to decode RGB encoded data
    '''
    data = data.astype(np.float64)
    return base + (((data[0] * 256 * 256) + (data[1] * 256) + data[2]) * interval)


def _range_check(datarange):
    '''
    Utility to check if data range is outside of precision for 3 digit base 256
    '''
    maxrange = 256 ** 3

    return datarange > maxrange





from __future__ import division
from rio_rgbify.encoders import data_to_rgb, _decode, _range_check
import numpy as np
import pytest


def test_encode_data_roundtrip():
    minrand, maxrand = np.sort(np.random.randint(-427, 8848, 2))

    testdata = np.round((np.sum(
        np.dstack(
            np.indices((512, 512),
                dtype=np.float64)),
        axis=2) / (511. + 511.)) * maxrand, 2) + minrand

    baseval = -1000
    interval = 0.1

    rtripped = _decode(data_to_rgb(testdata.copy(), baseval, interval), baseval, interval)

    assert testdata.min() == rtripped.min()
    assert testdata.max() == rtripped.max()


def test_encode_failrange():
    testdata = np.zeros((2))

    testdata[1] = 256 ** 3 + 1

    with pytest.raises(ValueError):
        data_to_rgb(testdata, 0, 1)


def test_catch_range():
    assert _range_check(256 ** 3 + 1)
    assert not _range_check(256 ** 3 - 1)




import click

import rasterio as rio
import numpy as np
from riomucho import RioMucho
import json
from rasterio.rio.options import creation_options

from rio_rgbify.encoders import data_to_rgb
from rio_rgbify.mbtiler import RGBTiler

def _rgb_worker(data, window, ij, g_args):
    return data_to_rgb(data[0][g_args['bidx'] - 1],
        g_args['base_val'],
        g_args['interval'])

@click.command('rgbify')
@click.argument('src_path', type=click.Path(exists=True))
@click.argument('dst_path', type=click.Path(exists=False))
@click.option('--base-val', '-b', type=float, default=0,
    help='The base value of which to base the output encoding on [DEFAULT=0]')
@click.option('--interval', '-i', type=float, default=1,
    help='Describes the precision of the output, by incrementing interval [DEFAULT=1]')
@click.option('--bidx', type=int, default=1,
    help='Band to encode [DEFAULT=1]')
@click.option('--max-z', type=int, default=None,
    help="Maximum zoom to tile (.mbtiles output only)")
@click.option('--bounding-tile', type=str, default=None,
    help="Bounding tile '[{x}, {y}, {z}]' to limit output tiles (.mbtiles output only)")
@click.option('--min-z', type=int, default=None,
    help="Minimum zoom to tile (.mbtiles output only)")
@click.option('--format', type=click.Choice(['png', 'webp']), default='png',
    help="Output tile format (.mbtiles output only)")
@click.option('--workers', '-j', type=int, default=4,
    help='Workers to run [DEFAULT=4]')
@click.option('--verbose', '-v', is_flag=True, default=False)
@click.pass_context
@creation_options
def rgbify(ctx, src_path, dst_path, base_val, interval, bidx, max_z, min_z, bounding_tile, format, workers, verbose, creation_options):
    if dst_path.split('.')[-1].lower() == 'tif':
        with rio.open(src_path) as src:
            meta = src.profile.copy()

        meta.update(
            count=3,
            dtype=np.uint8
        )

        for c in creation_options:
            meta[c] = creation_options[c]

        gargs = {
            'interval': interval,
            'base_val': base_val,
            'bidx': bidx
        }

        with RioMucho([src_path], dst_path, _rgb_worker,
            options=meta,
            global_args=gargs) as rm:

            rm.run(workers)

    elif dst_path.split('.')[-1].lower() == 'mbtiles':
        if min_z == None or max_z == None:
            raise ValueError('Zoom range must be provided for mbtile output')

        if max_z < min_z:
            raise ValueError('Max zoom {0} must be greater than min zoom {1}'.format(max_z, min_z))

        if bounding_tile is not None:
            try:
                bounding_tile = json.loads(bounding_tile)
            except:
                raise TypeError("Bounding tile of {0} is not valid".format(bounding_tile))

        with RGBTiler(src_path, dst_path,
                      interval=interval,
                      base_val=base_val,
                      format=format,
                      bounding_tile=bounding_tile,
                      max_z=max_z, min_z=min_z) as tiler:

            tiler.run(workers)

    else:
        raise ValueError('{} output filetype not supported'.format(dst_path.split('.')[-1]))
